#
# Copyright (c) 2015-2016 LAAS/CNRS
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software for any purpose
# with or without   fee is hereby granted, provided   that the above  copyright
# notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS  SOFTWARE INCLUDING ALL  IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR  BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR  ANY DAMAGES WHATSOEVER RESULTING  FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION,   ARISING OUT OF OR IN    CONNECTION WITH THE USE   OR
# PERFORMANCE OF THIS SOFTWARE.
#
#                                             Anthony Mallet on Wed May 13 2015
#
import collections
import functools
import json
import os
import threading

import actionlib
import genpy
import rospy
import rosservice
import xmlrpclib

import loader
import server
import session
from util import *

_master = None
_nodemap = {}

_requests = {}
_requests_id = 0


# --- init -----------------------------------------------------------------
#
def init():
  global _master

  rospy.init_node('rosix', anonymous=True)
  _master = xmlrpclib.ServerProxy(os.environ['ROS_MASTER_URI'])


# --- fini -----------------------------------------------------------------
#
def fini():
  global _nodemap

  for name, n in _nodemap.items():
    session.broadcast('unload', name)
  _nodemap = {}


# --- run ------------------------------------------------------------------
#
def run():
  rospy.spin()


# --- load -----------------------------------------------------------------
#
def load(name, args):
  global _master, _infomap

  # lookup node
  code, msg, val = _master.lookupNode('rosix', name)
  if code != 1:
    raise softerror(json.dumps(msg))

  # get published topics/services and connections
  code, msg, val = _master.getSystemState('rosix')
  if code == 1:
    pubs, subs, srvs = val
  else:
    raise softerror(json.dumps(msg))

  # filter out the looked up node
  pubs = [p[0] for p in pubs if "/" + name in p[1]]
  actions = [p[:-9] for p in pubs if p.endswith("/feedback")]
  topics = [p for p in pubs if '/'.join(p.split('/')[:-1]) not in actions]
  services = [s[0] for s in srvs if "/" + name in s[1]]

  # load topics class
  code, msg, ttypes = _master.getTopicTypes('rosix')
  if code != 1:
    raise softerror(json.dumps(msg))

  topicm = {}
  for p in topics:
    t = next(t for n,t in ttypes if n == p)
    c = loader.get_message_class(t)
    dat = collections.OrderedDict();
    for n, type in zip(c.__slots__, c._slot_types):
      dat[n] = _type2dict(type)

    pk = '/'.join([i for i in p.split('/')[1:] if i != name])
    topicm[pk] = {
      'name': p, 'type': t, 'class': c, 'data': dat
    }

  # load actions class
  actionm = {}
  for a in actions:
    t = next(t for n,t in ttypes if n == a + "/goal")
    t = t[:-4]
    c = loader.get_message_class(t)

    goal = loader.get_message_class(c._slot_types[0])
    goal = loader.get_message_class(goal._slot_types[2])
    inp = collections.OrderedDict();
    for n, type in zip(goal.__slots__, goal._slot_types):
      inp[n] = _type2dict(type)

    result = loader.get_message_class(c._slot_types[1])
    result = loader.get_message_class(result._slot_types[2])
    out = collections.OrderedDict();
    for n, type in zip(result.__slots__, result._slot_types):
      out[n] = _type2dict(type)

    ak = '/'.join([i for i in a.split('/')[1:] if i != name])
    actionm[ak] = {
      'name': a, 'type': t, 'class': c, 'inclass': goal,
      'input': inp, 'output': out
    }

  # load service class
  servicem = {}
  for s in services:
    t = rosservice.get_service_type(s)
    c = loader.get_service_class(t)

    inp = collections.OrderedDict();
    for n, type in zip(c._request_class.__slots__,
                          c._request_class._slot_types):
      inp[n] = _type2dict(type)

    out = collections.OrderedDict();
    for n, type in zip(c._response_class.__slots__,
                          c._response_class._slot_types):
      out[n] = _type2dict(type)

    sk = '/'.join([i for i in s.split('/')[1:] if i != name])
    servicem[sk] = {
      'name': s, 'type': t, 'class': c, 'input': inp, 'output': out
    }


  # register
  n = node_proxy(name, topicm, actionm, servicem)
  _nodemap[name] = n

  return name


# --- get_proxy ------------------------------------------------------------
#
def get_proxy(name):
  global _nodemap

  if name in _nodemap: return _nodemap[name]
  return None


# --- _type2dict -----------------------------------------------------------
#
def _type2dict(type):
  if type == 'bool': return {'kind':'bool'}
  elif type == 'byte': return {'kind':'octet'}
  elif type == 'uint8': return {'kind':'unsigned char'}
  elif type == 'uint16': return {'kind':'unsigned short'}
  elif type == 'uint32': return {'kind':'unsigned long'}
  elif type == 'string': return {'kind':'string'}
  elif type == 'float32': return {'kind':'float'}
  elif type == 'float64': return {'kind':'double'}

  elif type == 'time':
    return {
      'kind':'struct', 'type': collections.OrderedDict([
        ('secs', {'kind': 'unsigned long'}),
        ('nsecs', {'kind': 'unsigned long'})])
    }

  elif type[-2:] == '[]':
    d = _type2dict(type[:-2])
    return {'kind': 'sequence', 'type': d}

  elif type[-1:] == ']':
    r = type.rfind('[')
    d = _type2dict(type[:r])
    return {'kind': 'array', 'length': int(type[r+1:-1]),  'type': d}

  else:
    try:
      d = collections.OrderedDict();
      c = loader.get_message_class(type)
      for s, t in zip(c.__slots__, c._slot_types):
        d[s] = _type2dict(t)

      return {'kind': 'struct', 'type': d, 'class': c}
    except Exception as e:
      raise softerror(json.dumps(str(e)))

  raise softerror(json.dumps('invalid message type: %s' % type))


# --- _msg2dict ------------------------------------------------------------
#
def _msg2dict(msg, dict):
  if dict['kind'] == 'struct':
    d = collections.OrderedDict()
    for k, v in dict['type'].items():
      i = getattr(msg, k, None)
      d[k] = _msg2dict(i, v)

  elif dict['kind'] == 'sequence':
    d = map(lambda i: _msg2dict(i, dict['type']), msg)

  elif dict['kind'] == 'array':
    d = map(lambda i: _msg2dict(i, dict['type']), msg)

  else:
    d = msg

  return d


# --- _dict2msg ------------------------------------------------------------
#
def _dict2msg(d, dict):
  if dict['kind'] == 'struct':
    m = {}
    for k, v in dict['type'].items():
      if k in d:
        m[k] = _dict2msg(d[k], v)
    msg = dict['class'](**m)

  elif dict['kind'] == 'sequence':
    msg = map(lambda i: _dict2msg(i, dict['type']), d)

  elif dict['kind'] == 'array':
    msg = map(lambda i: _dict2msg(i, dict['type']), d)

  elif dict['kind'] == 'time':
    msg = genpy.Time(d['secs'], d['nsecs'])

  else:
    msg = d

  return msg


# --- node_proxy -----------------------------------------------------------
#
class node_proxy:
  keys = {}

  def __init__(self, name, topics, actions, services):
    self.name = name
    self.topics = topics
    self.actions = actions
    self.services = services

    for s, meta in services.items():
      self.services[s]['proxy'] = \
        rospy.ServiceProxy(meta['name'], meta['class'], True)

    for s, meta in actions.items():
      self.actions[s]['proxy'] = \
        actionlib.ActionClient(meta['name'], meta['class'])

  def __del__(self):
    global _requests

    if _requests is not None:
      for rid, r in _requests.items():
        if 'name' in r and r['name'] == self.name:
          clean(rid)

    for s, meta in self.services.items():
      meta['proxy'].close()


  def info(self):
    info = {
      'name': self.name, 'instance': self.name,
      'services': {}, 'ports': {}
    }

    for n, srv in self.services.items():
      info['services'][n] = {
        'input': dict_skip(srv['input'], 'class'),
        'output': dict_skip(srv['output'], 'class'),
        'meta': {}
      }

    for n, act in self.actions.items():
      info['services'][n] = {
        'input': dict_skip(act['input'], 'class'),
        'output': dict_skip(act['output'], 'class'),
        'meta': {}
      }

    for n, top in self.topics.items():
      info['ports'][n] = { 'simple': 1,
                           'data': dict_skip(top['data'], 'class'),
                           'meta': {} }

    return json.dumps(info)

  keys['info'] = info


  # --- send ---------------------------------------------------------------
  #
  def send(self, name, inp, args):
    global _requests, _requests_id

    if name in self.services:
      meta = self.services[name]
    elif name in self.actions:
      meta = self.actions[name]
    else:
      raise softerror(json.dumps('no such service: %s' % name))

    oneway = 'oneway' in args
    rid = _requests_id
    _requests_id += 1

    d = {}
    for k, v in meta['input'].items():
      if k in inp:
        d[k] = _dict2msg(inp[k], v)

    if name in self.services:
      i = meta['class']._request_class(**d)
      _requests[rid] = {
        'status': 'none',
        'node': self.name,
        'name': name,
        'channel': session.get_channel(args),
        't': threading.Thread(
          target=self.svc_call, args = [oneway, rid, name, i])
      }
      _requests[rid]['t'].start()

    else:
      i = meta['inclass'](**d)
      f = functools.partial(self.act_call_cb, oneway, rid)
      gh = self.actions[name]['proxy'].send_goal(i, f)
      gh.rid = rid
      _requests[rid] = {
        'status': 'none',
        'node': self.name,
        'name': name,
        'channel': session.get_channel(args),
        'gh': gh
      }

    return '' if oneway else str(rid)

  def svc_call(self, oneway, rid, name, inp):
    global _requests

    _requests[rid]['status'] = 'sent'
    if not oneway:
      channel = _requests[rid]['channel']
      if channel is not None:
        data = { 'id': '%s::%s' % (self.name, rid),
                 'status': 'sent' }
        server.push(channel, 'request', json.dumps(data))

    try:
      out = self.services[name]['proxy'](inp)

      if _requests[rid]['status'] != 'sent': return

      if not oneway:
        d = collections.OrderedDict()
        for k, v in self.services[name]['output'].items():
          d[k] = _msg2dict(getattr(out, k, None), v)

        result = { 'id': '%s::%s' % (self.name, rid),
                   'status': 'done', 'output': d }
        _requests[rid]['result'] = result
        _requests[rid]['status'] = 'done'
        if channel is not None:
          server.push(channel, 'request', json.dumps(result))
      else:
        self.clean(rid)

    except Exception as e:
      if not oneway:
        _requests[rid]['status'] = 'error'
        if channel is not None:
          data = { 'id': '%s::%s' % (self.name, rid),
                   'status': 'error', 'errmsg': str(e) }
          server.push(channel, 'request', json.dumps(data))
      else:
        self.clean(rid)

  def act_call_cb(self, oneway, rid, gh):
    global _requests

    channel = _requests[rid]['channel']
    s = gh.get_comm_state()

    if s == actionlib.CommState.ACTIVE:
      _requests[rid]['status'] = 'sent'
      if not oneway and channel is not None:
        data = { 'id': '%s::%s' % (self.name, rid), 'status': 'sent' }
        server.push(channel, 'request', json.dumps(data))

    elif s == actionlib.CommState.DONE:
      if _requests[rid]['status'] != 'sent': return

      if not oneway:
        out = gh.get_result()
        d = collections.OrderedDict()
        for k, v in self.actions[_requests[rid]['name']]['output'].items():
          d[k] = _msg2dict(getattr(out, k, None), v)

        result = { 'id': '%s::%s' % (self.name, rid),
                   'status': 'done', 'output': d }
        _requests[rid]['result'] = result
        _requests[rid]['status'] = 'done'
        if channel is not None:
          server.push(channel, 'request', json.dumps(result))
      else:
        self.clean(rid)


  # --- read ---------------------------------------------------------------
  #
  def read(self, name):
    if name not in self.topics:
      raise softerror(json.dumps('no such topic: %s' % name))

    if 'subscribed' not in self.topics[name]:
      rospy.Subscriber(self.topics[name]['name'],
                       self.topics[name]['class'], self.read_callback, name)
      self.topics[name]['subscribed'] = threading.Event()
      if not self.topics[name]['subscribed'].wait(3):
        raise softerror(json.dumps('timeout reading %s' % name))

    return self.topics[name]['payload']

  def read_callback(self, msg, name):
    d = collections.OrderedDict()
    for k, v in self.topics[name]['data'].items():
      d[k] = _msg2dict(getattr(msg, k, None), v)

    self.topics[name]['payload'] = json.dumps(d)
    self.topics[name]['subscribed'].set()


  # --- abort --------------------------------------------------------------
  #
  def abort(self, rid):
    global _requests

    if rid not in _requests:
      raise softerror(json.dumps('activity %d does not exist' % rid))
    if 'gh' in _requests[rid]:
      _requests[rid]['gh'].cancel()

    return ''


  # --- clean --------------------------------------------------------------
  #
  def clean(self, rid):
    global _requests

    if rid not in _requests:
      raise softerror(json.dumps('activity %d does not exist' % rid))

    del(_requests[rid])
    _requests[rid] = {'status': 'none'}
    return ''


# --- dict_skip ------------------------------------------------------------
#
def dict_skip(d, skip):
  if isinstance(d, collections.OrderedDict):
    return collections.OrderedDict((k, dict_skip(v, skip))
                                   for k,v in d.iteritems() if k != skip)

  elif isinstance(d, dict):
    return {k:dict_skip(v, skip) for k,v in d.iteritems() if k != skip}

  else:
    return d
